/* Programmieren mit Threads:
bell() ist ein Timer, der, 1 bis NUMmal hintereinander aufgerufen, 
jeweils zu den Weckzeiten tim.tBell[NUM] klingelt(alle Zeitangaben 
ab tim.start, in sec). Bilden die Weckzeiten keine steigende Folge 
(z.B.: ..,15,8,..), so werden versaeumte Termine nur mit je einem 
zusaetzlichen Ton verspaetet bedacht. Die Version MIT_THREADS erzeugt 
auch bei ungeordneten Weckzeiten Weck-Signale zum richtigen Zeitpunkt; 
MIT_CS beruecksichtigt zudem krit. Abschnitte und verhindert, dass 
mehrere Threads gleichzeitig im Konsole-Fenster ausgeben.
*/

#define MIT_THREADS
#ifdef  MIT_THREADS
// #define MIT_CS
#endif  MIT_THREADS

#include <conio.h>    /* getch() */
#include <process.h>  /* exit, _beginthread, _endthread */
#include <stdio.h>    /* printf */
#include <time.h>     /* time */
#ifdef  MIT_THREADS
 #include <windows.h> /* CRITICAL_SECTION */
#endif  MIT_THREADS

#define CLS "cls"
#define ESC  27
#define NUM   5

typedef struct /*Datentyp:*/
{ time_t start, tBell[NUM];
#ifdef  MIT_CS
  CRITICAL_SECTION cs4pr;
#endif  MIT_CS
} timerDat;    /* Typname)*/

/* *********************************************************************** */
   timerDat *init(void) /* Gemeinsamer Speicher: */
/* *********************************************************************** */
{ static timerDat tim;
  tim.tBell[0]=5;  tim.tBell[1]=10;  tim.tBell[2]=15;
  tim.tBell[3]=8;  tim.tBell[4]=7;
  time(&tim.start); 
  return(&tim);
}

/* *********************************************************************** */
   void bell(void *idx)
/* *********************************************************************** */
{
  static timerDat *tim=NULL;
  time_t tj=0, tTick=0;
  int j1=0;
  if (!tim) tim = init();

  do
  { time(&tj);  tj -= tim->start;

	  while (tj >= tTick)
	  { 
#ifdef  MIT_CS
      EnterCriticalSection (&tim->cs4pr); 
#endif  MIT_CS
      system(CLS);
      printf ("\n\r       [Idx:%d]   Zeit:%3d  [Weckzeiten:", (int)idx, tj);
      for (j1=0; j1<NUM; j1++) printf ("%4d", tim->tBell[j1]);
      printf ("]\r");
#ifdef  MIT_CS
      LeaveCriticalSection (&tim->cs4pr); 
#endif  MIT_CS
	    tTick++;
	  }
  } while (tj < tim->tBell[(int)idx]); 
  printf("\a"); /*beep*/
#ifdef  MIT_THREADS
  _endthread(); // wg. return entbehrlich
#endif  MIT_THREADS

  return;
}

/* *********************************************************************** */
   void multiBell(void) /* Beispielhafte Anwendung: */
/* *********************************************************************** */
{ int j1=0;
#ifdef  MIT_THREADS
  unsigned long hThread[NUM];

#ifdef  MIT_CS
  static timerDat *tim=NULL;
  if (!tim) tim = init();
  InitializeCriticalSection (&tim->cs4pr); 
#endif  MIT_CS
#endif  MIT_THREADS

  for (j1=0; j1<NUM; j1++) 
  {
#ifdef  MIT_THREADS
    /* int (o.a.) direkt zu void* casten: */
    hThread[j1] = _beginthread(bell, 0, (void *)j1); 
#else   MIT_THREADS
    bell((void *)j1);
#endif  MIT_THREADS
  }

#ifdef  MIT_THREADS //Conditio sine qua non (NUM=j1):
  WaitForMultipleObjects(NUM, (CONST HANDLE)hThread, TRUE, INFINITE);

#ifdef  MIT_CS
  DeleteCriticalSection (&tim->cs4pr); 
#endif  MIT_CS
#endif  MIT_THREADS
  return;
}

/* *********************************************************************** */
   int main (void) /* Zu Testzwecken: */
/* *********************************************************************** */
{ printf ("\n\r       Aris Christidis:");
  printf ("\n\r       Programm zum mehrfachen Wecken");
  printf ("\n\n       (mit bel. Taste weiter)\n");  
 _getch();  system (CLS); 

  multiBell();

  system (CLS); 
  printf ("\n\r       Beendigung mit bel. Taste\n\r       ");
 _getch(); 
  return 0;
}
